#include "fxlib.h"
#include "LibBMP.h"
#include <stdio.h>

Bitmap *BMP_Load(const char *adresse, int *err)
{
  unsigned char header[18];
  unsigned char *imgh;
  FONTCHARACTER *adr;
  Bitmap *b;
  int handle;
  int x;

  adr = calloc((strlen(adresse)+1),sizeof(FONTCHARACTER));
  for(x=0;x<strlen(adresse);x++) *(adr+x) = *(adresse+x);

  if((handle=Bfile_OpenFile(adr,_OPENMODE_READ))<0) { if(err) *err=handle; return NULL; }
  if(!(b=malloc(sizeof(Bitmap)))) { if(err) *err=1; Bfile_CloseFile(handle); return NULL; }
  if((x=Bfile_ReadFile(handle,header,18,0))<0) { if(err) *err=x; Bfile_CloseFile(handle); return NULL; }

  x = (header[17]<<24)+(header[16]<<16)+(header[15]<<8)+header[14];
  if(!(imgh=malloc(x))) { if(err) *err=1; Bfile_CloseFile(handle); return NULL; }
  if((x=Bfile_ReadFile(handle,imgh,x,18))<0) { if(err) *err=x; free(imgh); Bfile_CloseFile(handle); return NULL; }

  b->width  = (imgh[3]<<24)+(imgh[2]<<16)+(imgh[1]<<8)+imgh[0];
  b->height = (imgh[7]<<24)+(imgh[6]<<16)+(imgh[5]<<8)+imgh[4];
  b->depth  = (imgh[11]<<8)+imgh[10];

  if(b->depth!=1) { if(err)*err=2; free(imgh); Bfile_CloseFile(handle); return NULL; }
  x = (header[5]<<24)+(header[4]<<16)+(header[3]<<8)+header[2];
  if(!(b->data=malloc(x))) { if(err) *err=2; free(imgh); Bfile_CloseFile(handle); return NULL; }

  x = (imgh[15]<<24)+(imgh[14]<<16)+(imgh[13]<<8)+imgh[12];
  if(x) { if(err) *err=3; free(imgh); Bfile_CloseFile(handle); return NULL; }

  x = (imgh[19]<<24)+(imgh[18]<<16)+(imgh[17]<<8)+imgh[16];
  if(!(b->data=malloc(x))) { if(err) *err=1; free(imgh); Bfile_CloseFile(handle); return NULL; }

  if((x=Bfile_ReadFile(handle,b->data,x,header[10]))<0) { if(err) *err=x; }
  else if(err) *err=0;

  free(imgh);
  Bfile_CloseFile(handle);
  return x<0?NULL:b;
}

void BMP_Draw(Bitmap *bmp, int xb, int yb, int m)
{
  int ox = ((bmp->width-1)>>3)+1;
  unsigned char v;
  int i,x,y, a,b,c;

  for(y=0;y<bmp->height;y++) for(i=0;i<ox;i++)
  {
    v = bmp->data[ox*y+i];
    for(x=0;x<8;x++) {

      if((i<<3)+7-x<bmp->width)
      {
        a = xb+(i<<3)+7-x;
        b = yb+bmp->height-y-1;
        c = v&1;

        switch(m) {
          case BMP_AND: if(c)  Bdisp_SetPoint_VRAM(a,b,0); break;
          case BMP_OR:  if(!c) Bdisp_SetPoint_VRAM(a,b,1); break; 
          case BMP_XOR: if(!c) Bdisp_SetPoint_VRAM(a,b,!Bdisp_GetPoint_VRAM(a,b)); break; }
      }
      v >>= 1; }
  }
}

int BMP_GetPixel(Bitmap *b, int x, int y)
{
  int px;
  x = ((x>>3)<<3)+(7-x&7);
  px = ((b->width-1)>>3)+1;
  return !((*(b->data+(b->height-1-y)*px+((x-(x&7))>>3))>>(x&7))&1);
}

void BMP_SetPixel(Bitmap *b, int x, int y, int v)
{
  unsigned char byte;
  int px;
  x = ((x>>3)<<3)+(7-x&7);
  px = ((b->width-1)>>3)+1;
  byte = *(b->data+(b->height-1-y)*px+((x-(x&7))>>3));

  if(v) byte &= ~(1<<x%8);
  else byte |= (1<<x%8);
  *(b->data+(b->height-1-y)*px+((x-(x&7))>>3)) = byte;
}

void *BMP_GetData(Bitmap *bmp, int *err)
{
  int px = ((bmp->width-1)>>3)+1;
  unsigned char *data = malloc(px*bmp->height);
  int x,y;

  if(!data) { if(err) *err=1; return NULL; }

  for(y=0;y<bmp->height;y++) for(x=0;x<px;x++)
    *(data+y*px+x) = ~*(bmp->data+(bmp->height-y-1)*px+x);

  if(err) *err = 0;
  return data;
}

Bitmap *BMP_Create(void *d, int w, int h, int *err)
{
  Bitmap *bmp = (Bitmap *)malloc(sizeof(Bitmap));
  unsigned char *data = (unsigned char *)d;
  int px = ((w-1)>>3)+1; 
  int x,y;

  if(!bmp) { if(err) *err=1; return NULL; }
  bmp->data = malloc(h*px);
  if(!bmp->data) { if(err) *err=1; return NULL; }

  for(y=0;y<h;y++) for(x=0;x<px;x++)
    *(bmp->data+y*px+x) = ~*(data+(h-y-1)*px+x);
  bmp->width = w;
  bmp->height = h;

  if(err) *err=0;
  return bmp;
}

void BMP_Save(const char *adresse, Bitmap *bmp, int *err)
{
  unsigned char filh[14] = {0};
  unsigned char imgh[40] = {0};
  unsigned char palh[8]  = {0};
  FONTCHARACTER *adr;
  int i, handle;

  int imagesize = bmp->height*(((bmp->width-1)>>3)+1);
  int filesize  = imagesize + 14 + 40 + 8;
  int offset    = 62;
  int imghsize  = 40;

  filh[0x00] = 0x42;
  filh[0x01] = 0x4D;
  filh[0x02] = (filesize&0x000000FF);
  filh[0x03] = (filesize&0x0000FF00)>>8;
  filh[0x04] = (filesize&0x00FF0000)>>16;
  filh[0x05] = (filesize&0xFF000000)>>24;
  // 0x????????
  filh[0x0A] = (offset&0x000000FF);
  filh[0x0B] = (offset&0x0000FF00)>>8;
  filh[0x0C] = (offset&0x00FF0000)>>16;
  filh[0x0D] = (offset&0xFF000000)>>24;

  imgh[0x00] = (imghsize&0x000000FF);
  imgh[0x01] = (imghsize&0x0000FF00)>>8;
  imgh[0x02] = (imghsize&0x00FF0000)>>16;
  imgh[0x03] = (imghsize&0xFF000000)>>24;
  imgh[0x04] = (bmp->width&0x000000FF);
  imgh[0x05] = (bmp->width&0x0000FF00)>>8;
  imgh[0x06] = (bmp->width&0x00FF0000)>>16;
  imgh[0x07] = (bmp->width&0xFF000000)>>24;
  imgh[0x08] = (bmp->height&0x000000FF);
  imgh[0x09] = (bmp->height&0x0000FF00)>>8;
  imgh[0x0A] = (bmp->height&0x00FF0000)>>16;
  imgh[0x0B] = (bmp->height&0xFF000000)>>24;
  imgh[0x0C] = 0x01;
  imgh[0x0D] = 0x00;
  imgh[0x0E] = 0x01;
  imgh[0x0F] = 0x00;
  imgh[0x10] = 0x00;
  imgh[0x11] = 0x00;
  imgh[0x12] = 0x00;
  imgh[0x13] = 0x00;
  imgh[0x14] = (imagesize&0x000000FF);
  imgh[0x15] = (imagesize&0x0000FF00)>>8;
  imgh[0x16] = (imagesize&0x00FF0000)>>16;
  imgh[0x17] = (imagesize&0xFF000000)>>24;
  imgh[0x18] = 0x88;
  imgh[0x19] = 0x13;
  imgh[0x1A] = 0x00;
  imgh[0x1B] = 0x00;
  imgh[0x1C] = 0x88;
  imgh[0x1D] = 0x13;
  imgh[0x1E] = 0x00;
  imgh[0x1F] = 0x00;
  imgh[0x20] = 0x02;
  imgh[0x21] = 0x00;
  imgh[0x22] = 0x00;
  imgh[0x23] = 0x00;
  imgh[0x24] = 0x02;
  imgh[0x25] = 0x00;
  imgh[0x26] = 0x00;
  imgh[0x27] = 0x00;

  palh[0x00] = 0x00;
  palh[0x01] = 0x00;
  palh[0x02] = 0x00;
  // 0x??
  palh[0x04] = 0xFF;
  palh[0x05] = 0xFF;
  palh[0x06] = 0xFF;
  // 0x??

  for(i=0;*(adresse+i);i++);
  adr = calloc(i,sizeof(FONTCHARACTER));
  if(!adr) { if(err) *err=1; return; }
  for(i=0;*(adresse+i);i++) *(adr+i) = *(adresse+i);

  Bfile_DeleteFile(adr);
  if((handle=Bfile_CreateFile(adr,filesize))<0) { if(err) *err=handle; free(adr); return; }
  if((handle=Bfile_OpenFile(adr,_OPENMODE_WRITE))<0) { if(err) *err=handle; free(adr); return; }
  Bfile_WriteFile(handle,filh,14);
  Bfile_WriteFile(handle,imgh,40);
  Bfile_WriteFile(handle,palh,8);
  Bfile_WriteFile(handle,bmp->data,imagesize);
  Bfile_CloseFile(handle);

  if(err) *err=0;
  free(adr);
}

void BMP_Free(Bitmap *bmp)
{
  free(bmp->data);
  free(bmp);
}

